"""Store common constants."""
ACTION_TIMEOUT = 60.0
CONF_DATA = "data"
CONF_ID = "id"
CONF_TRIGGER = "trigger"
CONFIG_DIR = "config_dir"
LOG_LEVEL = "log_level"

# events
BASE_EVENT = "base_event"
CAMACQ_START_EVENT = "camacq_start_event"
CAMACQ_STOP_EVENT = "camacq_stop_event"
COMMAND_EVENT = "command_event"
IMAGE_EVENT = "image_event"
START_COMMAND_EVENT = "start_command_event"
STOP_COMMAND_EVENT = "stop_command_event"



"""Handle images."""
import logging
from collections import defaultdict

import numpy as np
import tifffile
import xmltodict

_LOGGER = logging.getLogger(__name__)


def read_image(path):
    """Read a tif image and return the data.

    Parameters
    ----------
    path : str
        The path to the image.

    Returns
    -------
    numpy array
        Return a numpy array with image data.
    """
    try:
        return tifffile.imread(path, key=0)
    except OSError as exception:
        _LOGGER.error("Bad path to image: %s", exception)
        return None


def save_image(path, data, description=None):
    """Save a tif image with image data and meta data.

    Parameters
    ----------
    path : str
        The path to the image.
    data : numpy array
        A numpy array with the image data.
    description : str
        The description string of the image.
    """
    tifffile.imsave(path, data, description=description)


def make_proj(images):
    """Make a dict of max projections from a dict of channels and paths.

    Each channel will make one max projection.

    Parameters
    ----------
    images : dict
        Dict of paths and channel ids.

    Returns
    -------
    dict
        Return a dict of channels that map ImageData objects.
        Each image object have a max projection as data.
    """
    _LOGGER.info("Making max projections...")
    sorted_images = defaultdict(list)
    max_imgs = {}
    for path, channel in images.items():
        image = ImageData(path=path)
        # Exclude images with 0, 16 or 256 pixel side.
        # pylint: disable=len-as-condition
        if len(image.data) == 0 or len(image.data) == 16 or len(image.data) == 256:
            continue
        sorted_images[channel].append(image)
        proj = np.max([img.data for img in sorted_images[channel]], axis=0)
        max_imgs[channel] = ImageData(path=path, data=proj, metadata=image.metadata)
    return max_imgs


class ImageData:
    """Represent the data of an image with path, data, metadata and histogram.

    Parameters
    ----------
    path : str
        Path to the image.
    data : numpy array
        A numpy array with the image data.
    metadata : dict
        The meta data of the image as a JSON dict.

    Attributes
    ----------
    path : str
        The path to the image.
    """

    # pylint: disable=too-many-arguments, too-many-instance-attributes

    def __init__(self, path=None, data=None, metadata=None):
        """Set up instance."""
        self.path = path
        self._data = data
        self.description = None
        if metadata is not None:
            self.metadata = metadata

    @property
    def data(self):
        """:numpy array: Return the data of the image.

        :setter: Set the data of the image.
        """
        if self._data is None:
            self._load_image_data()
        return self._data

    @data.setter
    def data(self, value):
        """Set the data of the image."""
        self._data = value

    @property
    def metadata(self):
        """:str: Return metadata of image.

        :setter: Set the meta data of the image.
        """
        if self.description is None:
            self._load_image_data()
        return xmltodict.parse(self.description)

    @metadata.setter
    def metadata(self, value):
        """Set the metadata of the image."""
        self.description = xmltodict.unparse(value)

    @property
    def histogram(self):
        """:numpy array: Calculate and return image histogram."""
        if self._data is None:
            self._load_image_data()
        if self._data.dtype.name == "uint16":
            max_int = 65535
        else:
            max_int = 255
        return np.histogram(self._data, bins=256, range=(0, max_int))

    def _load_image_data(self):
        """Load image data from path."""
        try:
            with tifffile.TiffFile(self.path) as tif:
                self._data = tif.asarray(key=0)
                self.description = tif.pages[0].description
        except (OSError, ValueError) as exception:
            _LOGGER.error("Bad path %s to image: %s", self.path, exception)

    def save(self, path=None, data=None, metadata=None):
        """Save image with image data and optional meta data.

        Parameters
        ----------
        path : str
            The path to the image.
        data : numpy array
            A numpy array with the image data.
        metadata : dict
            The meta data of the image as a JSON dict.
        """
        if path is None:
            path = self.path
        if data is None:
            data = self.data
        if metadata is None:
            metadata = self.metadata
        description = xmltodict.unparse(metadata)
        save_image(path, data, description)

    def __repr__(self):
        """Return the representation."""
        return f"ImageData(path={self.path})"



"""Handle templates."""
import jinja2
from jinja2.sandbox import ImmutableSandboxedEnvironment

from camacq.exceptions import TemplateError
from camacq.plugins.leica.sample import next_well_xy
from camacq.plugins.sample import get_matched_samples

TEMPLATE_ENV_DATA = "template_env"


def get_env(center):
    """Get the template environment."""
    if TEMPLATE_ENV_DATA not in center.data:
        env = ImmutableSandboxedEnvironment()
        env = _set_global(env, "next_well_xy", template_next_well_xy)
        env = _set_global(env, "next_well_x", template_next_well_x)
        env = _set_global(env, "next_well_y", template_next_well_y)
        env = _set_global(env, "matched_samples", get_matched_samples)
        center.data[TEMPLATE_ENV_DATA] = env
    return center.data[TEMPLATE_ENV_DATA]


def _set_global(env, func_name, func):
    """Set a template environment global function."""
    env.globals[func_name] = func
    return env


def make_template(center, data):
    """Make templated data."""
    if isinstance(data, dict):
        return {key: make_template(center, val) for key, val in data.items()}

    if isinstance(data, list):
        return [make_template(center, val) for val in data]

    env = get_env(center)
    return env.from_string(str(data))


def render_template(data, variables):
    """Render templated data."""
    if isinstance(data, dict):
        return {key: render_template(val, variables) for key, val in data.items()}

    if isinstance(data, list):
        return [render_template(val, variables) for val in data]

    try:
        rendered = data.render(variables)
    except jinja2.TemplateError as exc:
        raise TemplateError(exc)
    return rendered


def template_next_well_xy(sample, plate_name, x_wells=12, y_wells=8):
    """Return the next not done well for the given plate x, y format."""
    return next_well_xy(sample, plate_name, x_wells, y_wells)


def template_next_well_x(sample, plate_name, x_wells=12, y_wells=8):
    """Return the next well x coordinate for the plate x, y format."""
    x_well, _ = next_well_xy(sample, plate_name, x_wells=x_wells, y_wells=y_wells)
    return x_well


def template_next_well_y(sample, plate_name, x_wells=12, y_wells=8):
    """Return the next well x coordinate for the plate x, y format."""
    _, y_well = next_well_xy(sample, plate_name, x_wells=x_wells, y_wells=y_wells)
    return y_well



"""Handle sample state."""
import asyncio
import json
import logging
from abc import ABC, abstractmethod

import voluptuous as vol

from camacq.event import Event
from camacq.exceptions import SampleError
from camacq.helper import BASE_ACTION_SCHEMA
from camacq.util import dotdict

_LOGGER = logging.getLogger(__name__)
SAMPLE_EVENT = "sample_event"
SAMPLE_IMAGE_SET_EVENT = "sample_image_set_event"

ACTION_SET_SAMPLE = "set_sample"
SET_SAMPLE_ACTION_SCHEMA = BASE_ACTION_SCHEMA.extend(
    {"sample_name": vol.Coerce(str)}, extra=vol.ALLOW_EXTRA
)
BASE_SET_SAMPLE_ACTION_SCHEMA = BASE_ACTION_SCHEMA.extend(
    {vol.Required("name"): vol.Coerce(str), "values": dict}
)

ACTION_TO_METHOD = {
    ACTION_SET_SAMPLE: {"method": "set_sample", "schema": SET_SAMPLE_ACTION_SCHEMA},
}


async def setup_module(center, config):
    """Set up sample module.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """

    async def handle_action(**kwargs):
        """Handle action call to add a state to the sample.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments. These will be passed to a
            method when an action is called.
        """
        action_id = kwargs.pop("action_id")
        method = ACTION_TO_METHOD[action_id]["method"]
        sample_name = kwargs.pop("sample_name", None)
        silent = kwargs.pop("silent", False)
        if sample_name:
            samples = [center.samples[sample_name]]
        else:
            samples = list(center.samples.values())
        tasks = []
        for sample in samples:
            try:
                kwargs = sample.set_sample_schema(kwargs)
            except vol.Invalid as exc:
                _LOGGER.log(
                    logging.DEBUG if silent else logging.ERROR,
                    "Invalid action call parameters %s: %s for action: %s.%s",
                    kwargs,
                    exc,
                    "sample",
                    action_id,
                )
                continue

            _LOGGER.debug(
                "Handle sample %s action %s: %s", sample.name, action_id, kwargs
            )
            tasks.append(center.create_task(getattr(sample, method)(**kwargs)))
        if tasks:
            await asyncio.wait(tasks)

    for action_id, options in ACTION_TO_METHOD.items():
        schema = options["schema"]
        center.actions.register("sample", action_id, handle_action, schema)


class Samples(dotdict):
    """Hold all samples."""

    # pylint: disable=too-few-public-methods

    def __getattr__(self, sample_name):
        """Get a sample by name."""
        try:
            return self[sample_name]
        except KeyError as exc:
            raise SampleError(f"Unable to get sample with name {sample_name}") from exc


def register_sample(center, sample):
    """Register sample."""
    sample.center = center
    sample.data = {}
    center.bus.register(sample.image_event_type, sample.on_image)
    center.samples[sample.name] = sample


class ImageContainer(ABC):
    """A container for images."""

    @property
    @abstractmethod
    def change_event(self):
        """:Event: Return an event class to fire on container change."""

    @property
    @abstractmethod
    def images(self):
        """:dict: Return a dict with all images for the container."""

    @property
    @abstractmethod
    def name(self):
        """:str: Return an identifying name for the container."""

    @property
    @abstractmethod
    def values(self):
        """:dict: Return a dict with the values set for the container."""


class Sample(ImageContainer, ABC):
    """Representation of the state of the sample."""

    center = None
    data = None

    @property
    @abstractmethod
    def image_event_type(self):
        """:str: Return the image event type to listen to for the sample."""

    @property
    @abstractmethod
    def name(self):
        """:str: Return the name of the sample."""

    @property
    @abstractmethod
    def set_sample_schema(self):
        """Return the validation schema of the set_sample method."""

    @abstractmethod
    async def on_image(self, center, event):
        """Handle image event for this sample."""

    def get_sample(self, name, **kwargs):
        """Get an image container of the sample.

        Parameters
        ----------
        name : str
            The name of the container type.
        **kwargs
            Arbitrary keyword arguments.
            These will be used to create the id string of the container.

        Returns
        -------
        ImageContainer instance
            Return the found ImageContainer instance.
        """
        id_string = json.dumps({"name": name, **kwargs})
        return self.data.get(id_string)

    async def set_sample(self, name, values=None, **kwargs):
        """Set an image container of the sample.

        Parameters
        ----------
        name : str
            The name of the container type.
        values : dict
            The optional values to set on the container.
        **kwargs
            Arbitrary keyword arguments.
            These will be used to create the id string of the container.

        Returns
        -------
        ImageContainer instance
            Return the ImageContainer instance that was updated.
        """
        id_string = json.dumps({"name": name, **kwargs})
        values = values or {}
        container = self.data.get(id_string)
        event = None

        if container is None:
            container = await self._set_sample(name, values, **kwargs)
            event_class = container.change_event
            event = event_class({"container": container})

        container.values.update(values)
        self.data[id_string] = container

        if name == "image":
            self.images[container.path] = container

        if not event and values:
            event_class = container.change_event
            event = event_class({"container": container})

        if event:
            await self.center.bus.notify(event)
        return container

    @abstractmethod
    async def _set_sample(self, name, values, **kwargs):
        """Set an image container of the sample.

        Parameters
        ----------
        name : str
            The name of the container type.
        values : dict
            The values to set on the container.
        **kwargs
            Arbitrary keyword arguments.

        Returns
        -------
        ImageContainer instance
            Return the ImageContainer instance that was updated.
        """


class Image(ImageContainer):
    """An image with path and position info."""

    def __init__(self, path, values=None, **kwargs):
        """Set up instance."""
        self._path = path
        self._values = values or {}
        for attr, val in kwargs.items():
            setattr(self, attr, val)

    def __repr__(self):
        """Return the representation."""
        return f"<Image(path={self.path}, values={self.values})>"

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return SampleImageSetEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the container."""
        return {self.path: self}

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "image"

    @property
    def path(self):
        """:str: Return the path of the image."""
        return self._path

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class SampleEvent(Event):
    """An event produced by a sample change event."""

    __slots__ = ()

    event_type = SAMPLE_EVENT

    @property
    def container(self):
        """:ImageContainer instance: Return the container instance of the event."""
        return self.data.get("container")

    @property
    def container_name(self):
        """:str: Return the container name of the event."""
        return self.container.name

    @property
    def images(self):
        """:dict: Return the container images of the event."""
        return self.container.images

    @property
    def values(self):
        """:dict: Return the container values of the event."""
        return self.container.values

    def __repr__(self):
        """Return the representation."""
        data = dict(container=self.container)
        return f"{type(self).__name__}(data={data})"


class SampleImageSetEvent(SampleEvent):
    """An event produced by a new image on the sample."""

    __slots__ = ()

    event_type = SAMPLE_IMAGE_SET_EVENT


def get_matched_samples(sample, name, attrs=None, values=None):
    """Return the sample items that match."""
    attrs = attrs or {}
    values = values or {}
    items = [
        cont
        for cont in sample.data.values()
        if cont.name == name
        and (
            not attrs
            or all(getattr(cont, attr, None) == val for attr, val in attrs.items())
        )
        and (
            not values
            or all(cont.values.get(key) == val for key, val in values.items())
        )
    ]
    return items



"""Handle automations."""
# Copyright 2013-2017 The Home Assistant Authors
# https://github.com/home-assistant/home-assistant/blob/master/LICENSE.md
# This file was modified by The Camacq Authors.
import logging
from collections import deque
from functools import partial

import voluptuous as vol

from camacq.exceptions import TemplateError
from camacq.helper import BASE_ACTION_SCHEMA, get_module, has_at_least_one_key
from camacq.helper.template import make_template, render_template
from camacq.const import CAMACQ_STOP_EVENT, CONF_DATA, CONF_ID

_LOGGER = logging.getLogger(__name__)

CONF_AUTOMATIONS = "automations"
CONF_ACTION = "action"
CONF_CONDITION = "condition"
CONF_CONDITIONS = "conditions"
CONF_NAME = "name"
CONF_TRIGGER = "trigger"
CONF_TYPE = "type"
ENABLED = "enabled"
NAME = "name"
ACTION_DELAY = "delay"
ACTION_TOGGLE = "toggle"
DATA_AUTOMATIONS = "automations"

TRIGGER_ACTION_SCHEMA = vol.Schema(
    [
        {
            vol.Required(CONF_TYPE): vol.Coerce(str),
            vol.Required(CONF_ID): vol.Coerce(str),
            vol.Optional(CONF_DATA, default={}): dict,
        }
    ],
)

CONDITION_SCHEMA = vol.All(
    has_at_least_one_key(CONF_TYPE, CONF_CONDITION),
    {
        # pylint: disable=no-value-for-parameter
        vol.Inclusive(CONF_TYPE, "condition"): vol.All(
            vol.Upper, vol.In(["AND", "OR"])
        ),
        vol.Inclusive(CONF_CONDITIONS, "condition"): [
            # pylint: disable=unnecessary-lambda
            lambda value: CONDITION_SCHEMA(value)
        ],
        vol.Exclusive(CONF_CONDITION, "condition"): vol.Coerce(str),
    },
)

CONFIG_SCHEMA = vol.Schema(
    [
        {
            vol.Required(CONF_NAME): vol.Coerce(str),
            vol.Required(CONF_TRIGGER): TRIGGER_ACTION_SCHEMA,
            vol.Required(CONF_ACTION): TRIGGER_ACTION_SCHEMA,
            vol.Optional(
                CONF_CONDITION, default={CONF_CONDITION: "true"}
            ): CONDITION_SCHEMA,
        }
    ]
)


async def setup_module(center, config):
    """Set up automations package.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """
    _process_automations(center, config)
    automations = center.data[DATA_AUTOMATIONS]

    async def handle_action(**kwargs):
        """Enable or disable an automation."""
        name = kwargs[NAME]
        automation = automations[name]
        enabled = kwargs.get(ENABLED, not automation.enabled)
        if enabled:
            automation.enable()
        else:
            automation.disable()

    toggle_action_schema = BASE_ACTION_SCHEMA.extend(
        {
            vol.Required(NAME): vol.All(vol.Coerce(str), vol.In(automations)),
            ENABLED: vol.Boolean(),  # pylint: disable=no-value-for-parameter
        }
    )

    # register action to enable/disable automation
    center.actions.register(
        "automations", ACTION_TOGGLE, handle_action, toggle_action_schema
    )


def _process_automations(center, config):
    """Process automations from config."""
    automations = center.data.setdefault(DATA_AUTOMATIONS, {})
    conf = config[CONF_AUTOMATIONS]
    for block in conf:
        name = block[CONF_NAME]
        _LOGGER.debug("Setting up automation %s", name)
        action_sequence = _get_actions(center, block[CONF_ACTION])
        cond_func = _process_condition(center, block[CONF_CONDITION])
        # use partial to get a function with args to call later
        attach_triggers = partial(_process_trigger, center, block[CONF_TRIGGER])
        automations[name] = Automation(
            center, name, attach_triggers, cond_func, action_sequence
        )


def _get_actions(center, config_block):
    """Return actions."""
    actions = (TemplateAction(center, action_conf) for action_conf in config_block)

    return ActionSequence(center, actions)


def _process_condition(center, config_block):
    """Return a function that parses the condition."""
    if CONF_TYPE in config_block:
        checks = []
        condition_type = config_block[CONF_TYPE]
        conditions = config_block[CONF_CONDITIONS]
        for cond in conditions:
            check = _process_condition(center, cond)
            checks.append(check)
        return make_checker(condition_type, checks)

    data = config_block[CONF_CONDITION]
    template = make_template(center, data)
    return partial(render_template, template)


def make_checker(condition_type, checks):
    """Return a function to check condition."""

    def check_condition(variables):
        """Return True if all or any condition(s) pass."""
        if condition_type.lower() == "and":
            return all(template_check(check(variables)) for check in checks)
        if condition_type.lower() == "or":
            return any(template_check(check(variables)) for check in checks)
        return False

    return check_condition


def template_check(value):
    """Check if a rendered template string equals true.

    If value is not a string, return value as is.
    """
    if isinstance(value, str):
        return value.lower() == "true"
    return value


def _process_trigger(center, config_block, trigger):
    """Process triggers for an automation."""
    remove_funcs = []

    for conf in config_block:
        trigger_id = conf[CONF_ID]
        trigger_type = conf[CONF_TYPE]
        trigger_mod = get_module(__name__, trigger_type)
        if not trigger_mod:
            continue
        _LOGGER.debug("Setting up trigger %s", trigger_id)

        remove = trigger_mod.handle_trigger(center, conf, trigger)
        if not remove:
            _LOGGER.error("Setting up trigger %s failed", trigger_id)
            continue

        remove_funcs.append(remove)

    if not remove_funcs:
        return None

    def remove_triggers():
        """Remove attached triggers."""
        for remove in remove_funcs:
            remove()

    return remove_triggers


class Automation:
    """Automation class."""

    # pylint: disable=too-many-arguments

    def __init__(
        self, center, name, attach_triggers, cond_func, action_sequence, enabled=True
    ):
        """Set up instance."""
        self._center = center
        self.name = name
        self.enabled = False
        self._action_sequence = action_sequence
        self._attach_triggers = attach_triggers
        self._detach_triggers = None
        self._cond_func = cond_func
        if enabled:
            self.enable()

    def __repr__(self):
        """Return the representation."""
        return (
            f"Automation(center={self._center}, name={self.name}, "
            f"attach_triggers={self._attach_triggers}, cond_func={self._cond_func}, "
            f"action_sequence={self._action_sequence}, enabled={self.enabled})"
        )

    def enable(self):
        """Enable automation."""
        if self.enabled:
            return
        self._detach_triggers = self._attach_triggers(self.trigger)
        self.enabled = True

    def disable(self):
        """Disable automation."""
        if not self.enabled:
            return
        if self._detach_triggers is not None:
            self._detach_triggers()
        self._detach_triggers = None
        self.enabled = False

    async def trigger(self, variables):
        """Run actions of this automation."""
        variables["samples"] = self._center.samples
        _LOGGER.debug("Triggered automation %s", self.name)
        try:
            cond = self._cond_func(variables)
        except TemplateError as exc:
            _LOGGER.error("Failed to render condition for %s: %s", self.name, exc)
            return
        if cond:
            _LOGGER.debug("Condition passed for %s", self.name)
            await self._action_sequence(variables)


class ActionSequence:
    """Represent a sequence of actions."""

    # pylint: disable=too-few-public-methods

    def __init__(self, center, actions):
        """Set up instance."""
        self._center = center
        self.actions = list(actions)  # copy to list to make sure it's a list

    async def __call__(self, variables):
        """Start action sequence."""
        waiting = deque(self.actions)
        while waiting:
            action = waiting.popleft()

            if action.action_type == "automations" and action.action_id == ACTION_DELAY:
                rendered_kwargs = action.render(variables)
                seconds = rendered_kwargs.get("seconds")
                self.delay(float(seconds), variables, waiting)

            else:
                _LOGGER.debug(
                    "Calling action %s.%s", action.action_type, action.action_id
                )
                await action(variables)

    def delay(self, seconds, variables, waiting):
        """Delay action sequence.

        Parameters
        ----------
        seconds : float
            A time interval to delay the pending action sequence.
        variables : dict
            A dict of template variables.
        """
        sequence = ActionSequence(self._center, waiting)
        callback = partial(self._center.create_task, sequence(variables))
        waiting.clear()
        _LOGGER.info("Action delay for %s seconds", seconds)
        callback = self._center.loop.call_later(seconds, callback)

        async def cancel_pending_actions(center, event):
            """Cancel pending actions."""
            callback.cancel()

        self._center.bus.register(CAMACQ_STOP_EVENT, cancel_pending_actions)


class TemplateAction:
    """Representation of an action with template data."""

    # pylint: disable=too-few-public-methods

    def __init__(self, center, action_conf):
        """Set up instance."""
        self._center = center
        self.action_id = action_conf[CONF_ID]
        self.action_type = action_conf[CONF_TYPE]
        action_data = action_conf[CONF_DATA]
        self.template = make_template(center, action_data)

    async def __call__(self, variables=None):
        """Execute action with optional template variables."""
        try:
            rendered = self.render(variables)
        except TemplateError:
            return
        await self._center.actions.call(self.action_type, self.action_id, **rendered)

    def render(self, variables):
        """Render the template with the kwargs for the action."""
        variables = variables or {}
        try:
            rendered = render_template(self.template, variables)
        except TemplateError as exc:
            _LOGGER.error(
                "Failed to render variables for %s.%s: %s",
                self.action_type,
                self.action_id,
                exc,
            )
            raise
        return rendered



"""Handle renaming of an image."""
import logging
from pathlib import Path

import voluptuous as vol

from camacq.helper import BASE_ACTION_SCHEMA, has_at_least_one_key

_LOGGER = logging.getLogger(__name__)
ACTION_RENAME_IMAGE = "rename_image"


async def setup_module(center, config):
    """Set up image rename plugin.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """

    async def handle_action(**kwargs):
        """Handle the action call to rename an image.

        Parameters
        ----------
        **kwargs
            Arbitrary keyword arguments. These will be passed to the
            action function when an action is called.
        """
        sample_name = kwargs["sample"]
        old_path = Path(kwargs["old_path"])
        new_path = kwargs.get("new_path")
        new_name = kwargs.get("new_name")

        if new_name:
            old_dir = old_path.parent
            new_path = old_dir / new_name

        new_path = Path(new_path)  # make sure new_path is a Path instance

        result = await center.add_executor_job(rename_image, old_path, new_path)
        if not result:
            return
        sample = center.samples[sample_name]
        image = sample.images.pop(old_path, None)
        if image is None:
            return
        image_attrs = image.__dict__.copy()
        image_attrs.pop("_path")
        image_attrs.pop("_values")
        await sample.set_sample(
            image.name, path=new_path, values=image.values, **image_attrs
        )

    rename_image_action_schema = vol.All(
        BASE_ACTION_SCHEMA.extend(
            {
                vol.Required("sample"): vol.All(
                    vol.Coerce(str), vol.In(center.samples)
                ),
                vol.Required("old_path"): vol.Coerce(str),
                vol.Exclusive("new_path", "new_file"): vol.Coerce(str),
                vol.Exclusive("new_name", "new_file"): vol.Coerce(str),
            }
        ),
        has_at_least_one_key("new_path", "new_name"),
    )

    center.actions.register(
        "rename_image", ACTION_RENAME_IMAGE, handle_action, rename_image_action_schema,
    )


def rename_image(old_path, new_path):
    """Rename image at old_path to new_path.

    Parameters
    ----------
    old_path : str
        The absolute path to the existing image.
    new_path : str
        The absolute path to the renamed image.

    """
    renamed = False
    try:
        old_path.replace(new_path)
    except FileNotFoundError as exc:
        _LOGGER.error("File not found: %s", exc)
    except OSError as exc:
        _LOGGER.error("Failed to rename image: %s", exc)
    else:
        renamed = True
    return renamed



"""Main module."""
import argparse
import logging
import sys
from pathlib import Path

import camacq.bootstrap as bootstrap
import camacq.config as config_util
from camacq.const import CONFIG_DIR, LOG_LEVEL
from camacq.util import asyncio_run


def check_dir_arg(path):
    """Check that argument is a directory."""
    # remove if not needed
    path = Path(path)
    if path.is_dir():
        return path.resolve()
    raise argparse.ArgumentTypeError(f"String {path} is not a path to a directory")


def check_log_level(loglevel):
    """Validate log level and return it if valid."""
    # assuming loglevel is bound to the string value obtained from the
    # command line argument. Convert to upper case to allow the user to
    # specify --log=DEBUG or --log=debug
    numeric_level = getattr(logging, loglevel.upper(), None)
    if not isinstance(numeric_level, int):
        raise argparse.ArgumentTypeError(f"String {loglevel} is not a valid log level")
    return numeric_level


def parse_command_line(args=None):
    """Parse the provided command line."""
    parser = argparse.ArgumentParser(
        description="Control microscope through client-server program."
    )
    parser.add_argument(
        "--log-level",
        dest=LOG_LEVEL,
        type=check_log_level,
        help="an option to specify lowest log level to log",
    )
    parser.add_argument(
        "-C",
        "--config",
        dest=CONFIG_DIR,
        type=check_dir_arg,
        default=config_util.get_default_config_dir(),
        help="the path to camacq configuration directory",
    )
    args = parser.parse_args(args=args)
    cmd_args_dict = vars(args)
    cmd_args_dict = {key: val for key, val in cmd_args_dict.items() if val}

    return cmd_args_dict


def ensure_config_path(config_dir):
    """Validate the configuration directory."""
    # Test if configuration directory exists
    if config_dir.is_dir():
        return
    try:
        config_dir.mkdir()
    except OSError:
        print(
            "Fatal Error: Unable to create default configuration "
            f"directory: {config_dir}"
        )
        sys.exit(1)


def ensure_config_file(config_dir):
    """Ensure configuration file exists."""
    config_path = config_util.ensure_config_exists(config_dir)

    if config_path is None:
        print("Error getting configuration path")
        sys.exit(1)

    return config_path


async def setup_and_start(config_file, cmd_args):
    """Set up app and start."""
    try:
        center = await bootstrap.setup_file(config_file, cmd_args)
    except Exception as exc:  # pylint: disable=broad-except
        print("An error occurred during setup:", exc)
        return 1
    return await center.start()


def main(args=None):
    """Run main function."""
    # Parse command line arguments
    cmd_args = parse_command_line(args=args)
    config_dir = cmd_args[CONFIG_DIR]
    ensure_config_path(config_dir)
    config_file = ensure_config_file(config_dir)
    exit_code = asyncio_run(setup_and_start(config_file, cmd_args))
    return exit_code


if __name__ == "__main__":
    main()



"""Provide sample implementation for leica microscope."""
import json

import voluptuous as vol

from camacq.const import IMAGE_EVENT
from camacq.plugins.sample import (
    BASE_SET_SAMPLE_ACTION_SCHEMA,
    Image,
    ImageContainer,
    Sample,
    SampleEvent,
    register_sample,
)

LEICA_SAMPLE_EVENT = "leica_sample_event"
CHANNEL_EVENT = "channel_event"
FIELD_EVENT = "field_event"
PLATE_EVENT = "plate_event"
WELL_EVENT = "well_event"
Z_SLICE_EVENT = "z_slice_event"

SET_PLATE_SCHEMA = BASE_SET_SAMPLE_ACTION_SCHEMA.extend(
    {vol.Required("name"): "plate", vol.Required("plate_name"): vol.Coerce(str)}
)

SET_WELL_SCHEMA = SET_PLATE_SCHEMA.extend(
    {
        vol.Required("name"): "well",
        vol.Required("well_x"): vol.Coerce(int),
        vol.Required("well_y"): vol.Coerce(int),
        "values": vol.Schema({"well_img_ok": vol.Coerce(bool)}, extra=vol.ALLOW_EXTRA),
    }
)

SET_FIELD_SCHEMA = SET_WELL_SCHEMA.extend(
    {
        vol.Required("name"): "field",
        vol.Required("field_x"): vol.Coerce(int),
        vol.Required("field_y"): vol.Coerce(int),
        "values": vol.Schema({"field_img_ok": vol.Coerce(bool)}, extra=vol.ALLOW_EXTRA),
    }
)

SET_CHANNEL_SCHEMA = SET_WELL_SCHEMA.extend(
    {
        vol.Required("name"): "channel",
        vol.Required("channel_id"): vol.Coerce(int),
        "values": vol.Schema({"gain": vol.Coerce(float)}, extra=vol.ALLOW_EXTRA),
    }
)

SET_Z_SLICE_SCHEMA = SET_WELL_SCHEMA.extend(
    {vol.Required("name"): "z_slice", vol.Required("z_slice_id"): vol.Coerce(int)}
)

SET_IMAGE_SCHEMA = SET_FIELD_SCHEMA.extend(
    {
        vol.Required("name"): "image",
        vol.Required("path"): vol.Coerce(str),
        vol.Required("channel_id"): vol.Coerce(int),
        vol.Required("z_slice_id"): vol.Coerce(int),
    }
)

SET_SAMPLE_SCHEMA = vol.Any(
    SET_PLATE_SCHEMA,
    SET_WELL_SCHEMA,
    SET_FIELD_SCHEMA,
    SET_Z_SLICE_SCHEMA,
    SET_CHANNEL_SCHEMA,
    SET_IMAGE_SCHEMA,
)


async def setup_module(center, config):
    """Set up sample module.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """
    sample = LeicaSample()
    register_sample(center, sample)


class LeicaSample(Sample):
    """Representation of the state of the sample.

    The sample can have plates, wells, fields, channels and images.

    Parameters
    ----------
    images : dict
        A dict of images of the sample.
    values : dict
        Optional dict of values.
    """

    def __init__(self, images=None, values=None):
        """Set up instance."""
        self._images = images or {}
        self._values = values or {}

    def __repr__(self):
        """Return the representation."""
        return f"LeicaSample(images={self._images}, values={self._values})"

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return LeicaSampleEvent

    @property
    def image_event_type(self):
        """:str: Return the image event type to listen to for the sample."""
        return IMAGE_EVENT

    @property
    def images(self):
        """:dict: Return a dict with all images for the container."""
        return self._images

    @property
    def name(self):
        """:str: Return the name of the sample."""
        return "leica"

    @property
    def set_sample_schema(self):
        """Return the validation schema of the set_sample method."""
        return SET_SAMPLE_SCHEMA

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values

    async def on_image(self, center, event):
        """Handle image event for this sample."""
        await self.set_sample(
            "image",
            path=event.path,
            plate_name=event.plate_name,
            well_x=event.well_x,
            well_y=event.well_y,
            field_x=event.field_x,
            field_y=event.field_y,
            z_slice_id=event.z_slice_id,
            channel_id=event.channel_id,
        )

    async def _set_sample(self, name, values, **kwargs):
        """Set an image container of the sample."""
        sample = None

        if name == "image":
            params = SET_FIELD_SCHEMA({"name": "field", **kwargs})
            await self.set_sample(**params)
            params = SET_Z_SLICE_SCHEMA({"name": "z_slice", **kwargs})
            await self.set_sample(**params)
            params = SET_CHANNEL_SCHEMA({"name": "channel", **kwargs})
            await self.set_sample(**params)

            sample = Image(values=values, **kwargs)

        if name == "field":
            params = SET_PLATE_SCHEMA({"name": "plate", **kwargs})
            await self.set_sample(**params)
            params = SET_WELL_SCHEMA({"name": "well", **kwargs})
            await self.set_sample(**params)

            sample = Field(self._images, values=values, **kwargs)

        if name == "channel":
            params = SET_PLATE_SCHEMA({"name": "plate", **kwargs})
            await self.set_sample(**params)
            params = SET_WELL_SCHEMA({"name": "well", **kwargs})
            await self.set_sample(**params)

            sample = Channel(self._images, values=values, **kwargs)

        if name == "z_slice":
            params = SET_PLATE_SCHEMA({"name": "plate", **kwargs})
            await self.set_sample(**params)
            params = SET_WELL_SCHEMA({"name": "well", **kwargs})
            await self.set_sample(**params)

            sample = ZSlice(self._images, values=values, **kwargs)

        if name == "well":
            params = SET_PLATE_SCHEMA({"name": "plate", **kwargs})
            await self.set_sample(**params)

            sample = Well(self._images, values=values, **kwargs)

        if name == "plate":
            sample = Plate(self._images, values=values, **kwargs)

        return sample


class Plate(ImageContainer):
    """A container for wells.

    Parameters
    ----------
    images : dict
        All the images of the sample.
    plate_name: str
        The name of the plate.
    values : dict
        Optional dict of values.

    Attributes
    ----------
    plate_name: str
        The name of the plate.
    """

    def __init__(self, images, plate_name, **kwargs):
        """Set up instance."""
        self._images = images
        self.plate_name = plate_name
        self._values = kwargs.pop("values", {})

    def __repr__(self):
        """Return the representation."""
        return (
            f"Plate(images={self._images}, plate_name={self.plate_name}, "
            f"values={self._values})"
        )

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return PlateEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the plate."""
        return {
            image.path: image
            for image in self._images.values()
            if image.plate_name == self.plate_name
        }

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "plate"

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class Well(Plate, ImageContainer):
    """A well within a plate with fields and channels.

    Parameters
    ----------
    images : dict
        All the images of the sample.
    well_x : int
        x coordinate of the well, minimum 0.
    well_y : int
        y coordinate of the well, minimum 0.
    plate_name: str
        The name of the plate.
    values : dict
        Optional dict of values.

    Attributes
    ----------
    well_x : int
        Number showing the x coordinate of the well, from 0.
    well_y : int
        Number showing the y coordinate of the well, from 0.
    """

    def __init__(self, images, well_x, well_y, **kwargs):
        """Set up instance."""
        self._images = images
        self.well_x = well_x
        self.well_y = well_y
        self._values = kwargs.pop("values", {})
        super().__init__(images, **kwargs)

    def __repr__(self):
        """Return the representation."""
        return (
            f"Well(images={self._images}, well_x={self.well_x}, "
            f"well_y={self.well_y}, values={self._values})"
        )

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return WellEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the well."""
        return {
            image.path: image
            for image in self._images.values()
            if image.plate_name == self.plate_name
            and image.well_x == self.well_x
            and image.well_y == self.well_y
        }

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "well"

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class Field(Well, ImageContainer):
    """A field within a well.

    Parameters
    ----------
    images : dict
        All the images of the sample.
    field_x : int
        Coordinate of field in x.
    field_y : int
        Coordinate of field in y.
    well_x : int
        x coordinate of the well, minimum 0.
    well_y : int
        y coordinate of the well, minimum 0.
    plate_name: str
        The name of the plate.
    values : dict
        Optional dict of values.

    Attributes
    ----------
    field_x : int
        Number showing the x coordinate of the field, from 0.
    field_y : int
        Number showing the y coordinate of the field, from 0.
    """

    def __init__(self, images, field_x, field_y, **kwargs):
        """Set up instance."""
        self._images = images
        self.field_x = field_x
        self.field_y = field_y
        self._values = kwargs.pop("values", {})
        super().__init__(images, **kwargs)

    def __repr__(self):
        """Return the representation."""
        return (
            f"Field(images={self._images}, field_x={self.field_x}, "
            f"field_y={self.field_y}, values={self._values})"
        )

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return FieldEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the field."""
        return {
            image.path: image
            for image in self._images.values()
            if image.plate_name == self.plate_name
            and image.well_x == self.well_x
            and image.well_y == self.well_y
            and image.field_x == self.field_x
            and image.field_y == self.field_y
        }

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "field"

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class Channel(Well, ImageContainer):
    """A channel with attributes.

    Parameters
    ----------
    images : dict
        All the images of the sample.
    channel_id : int
        ID of the channel.
    well_x : int
        x coordinate of the well, minimum 0.
    well_y : int
        y coordinate of the well, minimum 0.
    plate_name: str
        The name of the plate.
    values : dict
        Optional dict of values.

    Attributes
    ----------
    channel_id : int
        Return channel_id of the channel.
    """

    def __init__(self, images, channel_id, **kwargs):
        """Set up instance."""
        self._images = images
        self.channel_id = channel_id
        self._values = kwargs.pop("values", {})
        super().__init__(images, **kwargs)

    def __repr__(self):
        """Return the representation."""
        return (
            f"Channel(images={self._images}, channel_id={self.channel_id}, "
            f"values={self._values})"
        )

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return ChannelEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the channel."""
        return {
            image.path: image
            for image in self._images.values()
            if image.plate_name == self.plate_name
            and image.well_x == self.well_x
            and image.well_y == self.well_y
            and image.channel_id == self.channel_id
        }

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "channel"

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class ZSlice(Well, ImageContainer):
    """A channel with attributes.

    Parameters
    ----------
    images : dict
        All the images of the sample.
    z_slice_id : int
        ID of the slice.
    well_x : int
        x coordinate of the well, minimum 0.
    well_y : int
        y coordinate of the well, minimum 0.
    plate_name: str
        The name of the plate.
    values : dict
        Optional dict of values.

    Attributes
    ----------
    z_slice_id : int
        Return z_slice_id of the channel.
    """

    def __init__(self, images, z_slice_id, **kwargs):
        """Set up instance."""
        self._images = images
        self.z_slice_id = z_slice_id
        self._values = kwargs.pop("values", {})
        super().__init__(images, **kwargs)

    def __repr__(self):
        """Return the representation."""
        return (
            f"ZSlice(images={self._images}, z_slice_id={self.z_slice_id}, "
            f"values={self._values})"
        )

    @property
    def change_event(self):
        """:Event: Return an event class to fire on container change."""
        return ZSliceEvent

    @property
    def images(self):
        """:dict: Return a dict with all images for the channel."""
        return {
            image.path: image
            for image in self._images.values()
            if image.plate_name == self.plate_name
            and image.well_x == self.well_x
            and image.well_y == self.well_y
            and image.z_slice_id == self.z_slice_id
        }

    @property
    def name(self):
        """:str: Return an identifying name for the container."""
        return "z_slice"

    @property
    def values(self):
        """:dict: Return a dict with the values set for the container."""
        return self._values


class LeicaSampleEvent(SampleEvent):
    """An event produced by a sample change event."""

    __slots__ = ()

    event_type = LEICA_SAMPLE_EVENT


class PlateEvent(LeicaSampleEvent):
    """An event produced by a sample plate change event."""

    __slots__ = ()

    event_type = PLATE_EVENT

    @property
    def plate_name(self):
        """:str: Return the name of the plate."""
        return self.container.plate_name


class WellEvent(PlateEvent):
    """An event produced by a sample well change event."""

    __slots__ = ()

    event_type = WELL_EVENT

    @property
    def well_x(self):
        """:int: Return the well x coordinate of the event."""
        return self.container.well_x

    @property
    def well_y(self):
        """:int: Return the well y coordinate of the event."""
        return self.container.well_y

    @property
    def well_img_ok(self):
        """:bool: Return if the well has all images acquired ok."""
        return self.container.values.get("well_img_ok", False)


class ChannelEvent(WellEvent):
    """An event produced by a sample channel change event."""

    __slots__ = ()

    event_type = CHANNEL_EVENT

    @property
    def channel_id(self):
        """:int: Return the channel id of the event."""
        return self.container.channel_id

    @property
    def channel_name(self):
        """:str: Return the channel name of the event."""
        return self.container.values.get("channel_name")


class FieldEvent(WellEvent):
    """An event produced by a sample field change event."""

    __slots__ = ()

    event_type = FIELD_EVENT

    @property
    def field_x(self):
        """:int: Return the field x coordinate of the event."""
        return self.container.field_x

    @property
    def field_y(self):
        """:int: Return the field y coordinate of the event."""
        return self.container.field_y

    @property
    def field_img_ok(self):
        """:bool: Return if the field has all images acquired ok."""
        return self.container.values.get("field_img_ok", False)


class ZSliceEvent(WellEvent):
    """An event produced by a sample z slice change event."""

    __slots__ = ()

    event_type = Z_SLICE_EVENT

    @property
    def z_slice_id(self):
        """:int: Return the z_slice id of the event."""
        return self.container.z_slice_id


def next_well_xy(sample, plate_name, x_wells=None, y_wells=None):
    """Return the next not done well for the given plate x, y format."""
    if json.dumps({"name": "plate", "plate_name": plate_name}) not in sample.data:
        return None, None
    if x_wells is None or y_wells is None:
        not_done = (
            (cont.well_x, cont.well_y)
            for cont in sample.data.values()
            if cont.name == "well"
            and cont.plate_name == plate_name
            and not cont.values.get("well_img_ok", False)
        )
    else:
        done = {
            (cont.well_x, cont.well_y)
            for cont in sample.data.values()
            if cont.name == "well"
            and cont.plate_name == plate_name
            and cont.values.get("well_img_ok", False)
        }
        not_done = (
            (x_well, y_well)
            for x_well in range(x_wells)
            for y_well in range(y_wells)
            if (x_well, y_well) not in done
        )

    x_well, y_well = next(not_done, (None, None))
    return x_well, y_well



"""Configure and set up control center."""
import logging

import camacq.config as config_util
import camacq.log as log_util
from camacq.control import Center
from camacq.helper import setup_one_module
from camacq import plugins

_LOGGER = logging.getLogger(__name__)


async def setup_dict(center, config):
    """Set up control center from config dict.

    Parameters
    ----------
    config : dict
        The config dict.

    Returns
    -------
    Center instance
        Return the Center instance.
    """
    log_util.enable_log(config)
    await setup_one_module(center, config, plugins)


async def setup_file(config_file, cmd_args):
    """Set up control center from config file and command line args.

    Parameters
    ----------
    config_file : str
        The path to the configuration YAML file.
    cmd_args : dict
        The dict with the command line arguments.

    Returns
    -------
    Center instance
        Return the Center instance.
    """
    center = Center()
    user_config = await center.add_executor_job(
        config_util.load_config_file, config_file
    )
    user_config.update(cmd_args)  # merge config dict with command line args
    await setup_dict(center, user_config)
    return center



"""Set up file for camacq package."""
from pathlib import Path

from setuptools import find_packages, setup

PROJECT_DIR = Path(__file__).parent.resolve()
VERSION = (PROJECT_DIR / "camacq" / "VERSION").read_text().strip()

GITHUB_URL = "https://github.com/CellProfiling/cam_acq"
REQUIRES = [
    "async_timeout",
    "colorlog",
    "jinja2",
    "leicacam>=0.4.0",
    "leicaimage",
    "numpy",
    "ruamel.yaml>=0.15",
    "tifffile",
    "voluptuous",
    "xmltodict",
]

README_FILE = PROJECT_DIR / "README.md"
LONG_DESCR = README_FILE.read_text(encoding="utf-8")

DOWNLOAD_URL = f"{GITHUB_URL}/archive/master.zip"
CLASSIFIERS = [
    # How mature is this project? Common values are
    #   3 - Alpha
    #   4 - Beta
    #   5 - Production/Stable
    "Development Status :: 3 - Alpha",
    # Indicate who your project is intended for
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
    # Pick your license as you wish (should match "license" above)
    "License :: OSI Approved :: Apache Software License",
    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
]

CONFIG = {
    "description": "Control microscope through client server program.",
    "long_description": LONG_DESCR,
    "long_description_content_type": "text/markdown",
    "author": "Martin Hjelmare",
    "url": GITHUB_URL,
    "download_url": DOWNLOAD_URL,
    "license": "Apache-2.0",
    "author_email": "marhje52@kth.se",
    "version": VERSION,
    "python_requires": ">=3.6",
    "install_requires": REQUIRES,
    "packages": find_packages(exclude=["contrib", "docs", "tests*"]),
    "include_package_data": True,
    "entry_points": {
        "console_scripts": ["camacq = camacq.__main__:main"],
        "camacq.plugins": [
            "api = camacq.plugins.api",
            "automations = camacq.plugins.automations",
            "leica = camacq.plugins.leica",
            "rename_image = camacq.plugins.rename_image",
            "sample = camacq.plugins.sample",
        ],
    },
    "name": "camacq",
    "zip_safe": False,
    "classifiers": CLASSIFIERS,
}

setup(**CONFIG)



"""Handle plugins."""
import asyncio

import pkg_resources

from camacq.helper import setup_one_module

CORE_MODULES = ["api", "sample"]


async def setup_module(center, config):
    """Set up the plugins package.

    Parameters
    ----------
    center : Center instance
        The Center instance.
    config : dict
        The config dict.
    """
    plugins = await center.add_executor_job(get_plugins)

    # Add core modules.
    tasks = []
    for module_name in CORE_MODULES:
        if module_name not in config:
            config[module_name] = {}
        module = plugins.pop(module_name, None)
        if not module:
            continue
        tasks.append(setup_one_module(center, config, module))
    if tasks:
        await asyncio.wait(tasks)

    tasks = []
    for name, module in plugins.items():
        if name not in config:
            continue
        tasks.append(setup_one_module(center, config, module))
    if tasks:
        await asyncio.wait(tasks)


def get_plugins():
    """Return a dict of plugin modules."""
    plugins = {
        entry_point.name: entry_point.load()
        for entry_point in pkg_resources.iter_entry_points("camacq.plugins")
    }
    return plugins
